-- Delay function courtesy of http://wowwiki.wikia.com/wiki/USERAPI_wait
local waitTable = {};
local waitFrame = nil;

-- Variables used to stop message received spammage
LWORadio.CurrentTime = 0;
LWORadio.LastTime    = 0;

local MsgTag = "LWOTest"

function LWORadio:DelayFunction(delay, func, ...)
  if(type(delay)~="number" or type(func)~="function") then
    return false;
  end
  if(waitFrame == nil) then
    waitFrame = CreateFrame("Frame","LWOWaitFrame", UIParent);
    waitFrame:SetScript("onUpdate",function (self,elapse)
      local count = #waitTable;
      local i = 1;
      while(i<=count) do
        local waitRecord = tremove(waitTable,i);
        local d = tremove(waitRecord,1);
        local f = tremove(waitRecord,1);
        local p = tremove(waitRecord,1);
        if(d>elapse) then
          tinsert(waitTable,i,{d-elapse,f,p});
          i = i + 1;
        else
          count = count - 1;
          f(unpack(p));
        end
      end
    end);
  end
  tinsert(waitTable,{delay,func,{...}});
  return true;
end;

-- Decimal colour to hexadecimal colour courtesy of http://wowprogramming.com/snippets/Convert_decimal_classcolor_into_hex_27
function LWORadio:DecimalToHex(r,g,b)
    local r = string.format("%02x", r * 255);
    local g = string.format("%02x", g * 255);
    local b = string.format("%02x", b * 255);
    
    return r, g, b
end;

-- Rounding function courtesy of http://wowwiki.wikia.com/wiki/USERAPI_round
function LWORadio:Round(number, decimals)
    return (("%%.%df"):format(decimals)):format(number)
end;

function LWORadio:IdentifyChatFrames()
    for i = 1, NUM_CHAT_WINDOWS do
        local Frame = _G["ChatFrame"..i]
        if Frame then
            Frame:AddMessage(string.format(LWORadio_Localisation.IDENTIFY_CHAT_FRAME, self.db.char["Name"], tostring(i)), self.db.char["MessageColour"][1], self.db.char["MessageColour"][2], self.db.char["MessageColour"][3]);
        end
    end
end;

function LWORadio:FetchPresetList()
    local PresetList = {}
    
    for k, v in pairs(LWORadio_PresetOptions) do
    
        PresetList[k] = v.char["Name"];
    
    end
    
    return PresetList;

end;

function LWORadio:FakeDistressMsg(CharName)
    local CharacterName;
    local SubZone;
    local Location;
    local Zone;
    
    Location = self:GetPlayerLocation();
    SubZone = GetSubZoneText();
    
    if string.len(SubZone) ~= 0 then
        Zone = SubZone .. ", " .. Location["Area"];
    else
        Zone = Location["Area"];
    end
        
    CharacterName = CharName
    
    local PanicMessage = string.format(LWORadio.db.char["PanicButtonMessage"], CharacterName, Zone, self:Round(Location["x"], 2) .. ", " .. self:Round(Location["y"], 2))
      
    self:SendMessage(PanicMessage, "PANIC", CharName);
    

end;

function LWORadio:SetPreset(PresetTable)
    for k,v in pairs(LWORadio.db.char) do
    
        LWORadio.db.char[k] = PresetTable[k];
    
    end

end;

function LWORadio:TextareaToCSV(InputText)
    local regex = "[^a-zA-Z0-9\n]";
    --local InputText; = string.gsub(InputText, regex, nil);
    -- Removes blank newline entries
    InputText = string.gsub(InputText, regex, "");
    InputText = string.gsub(InputText, "[\n]+", ",");
    
    if InputText == "," then
        InputText = "";
    end
    
    -- Get end character of CSV list to check if it's a comma too
    local EndChar = string.sub(InputText, string.len(InputText));
    
    if EndChar == "," then
        return string.sub(InputText, 1, string.len(InputText) -1);
    end
    
    return InputText;
    
end;

function LWORadio:CSVToTable(InputText)
    local tmp = { strsplit(",", InputText) };

    return tmp;

end;

function LWORadio:CSVToTextarea(InputText)
    local tmp = string.gsub(InputText, ",", "\n");
    return tmp;

end;

function LWORadio:SetDeviceEmoteNoises(val)
    local EmoteNoise = self:TextareaToCSV(val);
    
    if EmoteNoise == "" then
        EmoteNoise = self.DefaultOptions.char["DeviceEmoteNoises"];
    end
    
    return EmoteNoise;

end;

-- true for table, false for string
function LWORadio:GetDeviceEmoteNoises(asTable)
    asTable = asTable or false;
    
    if asTable == false then
        return self:CSVToTextarea(self.db.char["DeviceEmoteNoises"]);
    else
        return self:CSVToTable(self.db.char["DeviceEmoteNoises"]);
    end

end;

-- Sets the channel prefix for addon communication message function
function LWORadio:SetCommFrequencies(NewPrefixList)
    local PrefixList = self:TextareaToCSV(NewPrefixList);
    
    if PrefixList == "" then
        PrefixList = LWORadio.DefaultOptions.char["ChannelPrefix"];
    end
    
    self.db.char["ChannelPrefix"] = PrefixList;

end;

function LWORadio:GetCommFrequencies(asTable)
    asTable = asTable or false;

    if asTable == false then
        return self:CSVToTextarea(self.db.char["ChannelPrefix"]);
    else
        return self:CSVToTable(self.db.char["ChannelPrefix"]);
    end

end;

function LWORadio:SetTransmitters(val)
    local tmp           = {};
    local firstStep     = { strsplit("\n", val) };
    local secondStep    = {};

    for k, v in pairs(firstStep) do
        tmp = { strsplit(",", v) };
        secondStep[k] = { ["x"] = tmp[1], ["y"] = tmp[2], ["Area"] = tmp[3] };
    end

    return secondStep;

end;

function LWORadio:GetTransmitters()
    local tmp = {};
    
    for k, v in pairs(self.db.char["Transmitters"]) do
        tmp[k] = string.format("%s,%s,%s", v["x"], v["y"], v["Area"]);
    end
    
    tmp = table.concat(tmp, "\n");
    
    return tmp;

end;

function LWORadio:GetClosestTransmitter(Location)
    Location    = Location or LWORadio:GetPlayerLocation();
    local TransmitterDistances  = {}
    local Transmitters          = self.db.char["Transmitters"];
    local DistanceCoords;   -- Distance in cartesian coordinates
    local DistanceYrds;     -- Distance in yards
    local AngleRads;
    local AngleDegs;
    local DistanceMeasurement;
    local ClosestTransmitter = { ["Distance"] = -1, ["DistanceYrds"] = -1, ["x"] = 0, ["y"] = 0 };
    local DistanceMeasurementKey     
    
    if Transmitters == false then 
    
        return false;
        
    end
    
    for k, v in pairs(Transmitters) do
        -- If the transmitter is in the same area as us, and the transmitters MAP translated to localisation is the same as the area we are in
        if ( string.lower(v["Area"]) == string.lower(Location["Area"]) ) --[[and ( string.lower(v["Area"]) == string.lower(LWOMapFiles:MapLocalize(Location["Map"])) )]] then
        
            DistanceCoords  = self:DistanceVector(v, Location, false);
            --DistanceYrds    = LWOMapFiles:DistanceAndDirection(Location["Map"], 1, v["x"]/100, v["y"]/100 );
            DistanceYrds    = LWOMapFiles:GetZoneDistance(Location["Map"], Location["x"]/100, Location["y"]/100, Location["Map"], v["x"]/100, v["y"]/100 );
            
                    if self.db.char["DistanceInYards"] == true then
                    
                        DistanceMeasurement     = DistanceYrds;
                        DistanceMeasurementKey  = "DistanceYrds";
                        
                    else
                    
                        DistanceMeasurement     = DistanceCoords;
                        DistanceMeasurementKey = "Distance";
                        
                    end
            
            -- If this transmitter is closer than the previous transmitter, and a previous transmitter has already been selected -- Or if no transmitter in this zone has been selected
            if ( DistanceMeasurement < ClosestTransmitter[DistanceMeasurementKey] and ClosestTransmitter[DistanceMeasurementKey] ~= -1 ) or ( ClosestTransmitter[DistanceMeasurementKey] == -1 ) then

                ClosestTransmitter = { 
                    ["Distance"]        = DistanceCoords, 
                    ["DistanceYrds"]    = DistanceYrds,
                    ["x"]               = v["x"], 
                    ["y"]               = v["y"], 
                    ["PlayerLocation"]  = Location 
                };
                
            end
            
        end
        
    end
    
    return ClosestTransmitter;
    
end;

function LWORadio:ShowPlayerClosestTransmitter()
    --math.atan2(y2-y1, x2-x1)
    local Transmitter       = self:GetClosestTransmitter();
    
    if Transmitter ~= false then
    
        local Distance;
        local PlayerLocation        = Transmitter.PlayerLocation;
        local TransmitterLocation   = string.format("X:%.2f, Y:%.2f", Transmitter["x"], Transmitter["y"] );
        
        if self.db.char["DistanceInYards"] == true then
        
            Distance = string.format("%.2f yards", Transmitter["DistanceYrds"]);
        
        else
        
            Distance = string.format("%.2f points", Transmitter["Distance"]);
        end

        local ChatFrame     = _G["ChatFrame" .. self.db.char["OutputChatFrame"]];

        ChatFrame:AddMessage(string.format(LWORadio_Localisation.NOTICE_CLOSEST_TRANSMITTER, self.db.char["Name"], Distance, TransmitterLocation), self.db.char["MessageColour"][1], self.db.char["MessageColour"][2], self.db.char["MessageColour"][3]);

    end

end;

function LWORadio:AddInterferenceToMessage(Message, Level)
    local InterferenceMessage = "";
    local MessageLength = string.len(Message);
    local StringSplits;
    local LetterEngulf;
    local tmp = {};
    local SplitInterval;
    local Pivot;
    
    -- Small amount of interference
    if Level == 0 then
    
        return Message;
    
    elseif Level == 1 then
    
        StringSplits    = math.random(0,3);     -- How many bits of interference are we going to be doing?
        LetterEngulf    = math.random(1,2);     -- How many letters do we engulf on each point for interference?
    
    -- Large amount of interference
    elseif Level == 2 then
    
        StringSplits = math.random(0,5);
        LetterEngulf = math.random(2,3);
    
    -- Huge amount of interference
    else
    
        StringSplits = math.random(3,8);
        LetterEngulf = math.random(3,6);
    
    end
    
    if StringSplits > MessageLength then
    
        StringSplits = MessageLength
        
    end
    
    SplitInterval   = math.ceil(MessageLength / StringSplits); -- Split the string up in to the number of parts specified by StringSplits

    if ((LetterEngulf * 2) + 1 ) * StringSplits < MessageLength then

        for i = 1, StringSplits, 1 do

            
            local StartPoint     = 1 + (SplitInterval * ( i - 1 ) );
            local EndPoint       = SplitInterval * i;
            
            if EndPoint > MessageLength then
            
                EndPoint = MessageLength;
                
            end
            
            Pivot = math.random(StartPoint+LetterEngulf, EndPoint-LetterEngulf);
            
            tmp[i] = { 
                ["StartPoint"]   = string.sub(Message, StartPoint, Pivot-LetterEngulf),
                ["EndPoint"]     = string.sub(Message, Pivot+LetterEngulf+1, EndPoint)
            }
            
        end
        
        for k, v in pairs(tmp) do
        
            InterferenceMessage = InterferenceMessage .. v["StartPoint"] .. " *static* " .. v["EndPoint"];
        
        end
        
    else
    
        InterferenceMessage = " *static* ";
        
    end
    
    if StringSplits == 0 then
    
        return Message;
        
    end
    
    return InterferenceMessage;

end;

-- Point a, Point b, give result squared or perform square root operatin
function LWORadio:DistanceVector(PointA, PointB, squared)
    local dX = PointA.x - PointB.x;
    local dY = PointA.y - PointB.y
    local vector = (dX * dX) + (dY * dY);
    
    squared = squared or false;
    
    if squared == true then
        return vector;
    end
    
    return math.sqrt(vector);

end;

function LWORadio:GetPlayerLocation()
    local uiMapID       = C_Map.GetBestMapForUnit("player");
    local Pos           = C_Map.GetPlayerMapPosition(uiMapID, "player");
    local Area          = GetZoneText();
    local Map           = C_Map.GetMapInfo(uiMapID).mapID;
    
    return { ["x"] = (Pos.x * 100), ["y"] = (Pos.y * 100), ["Area"] = Area, ["Map"] = Map };

end;

function LWORadio:RegisterAddonChannel()
    local CommFrequencies = self:GetCommFrequencies(true);
    
    for k, v in pairs(CommFrequencies) do
        self:RegisterComm(v, LWORadio.ReceiveMessage);
    end

end;

function LWORadio:GetTextGender()
    return self.Gender[self.PlayerGender];

end;

function LWORadio:ToggleRadioState(info, val)
    local ChatFrame     = _G["ChatFrame" .. self.db.char["OutputChatFrame"]];

    if val == true then
        ChatFrame:AddMessage(string.format(LWORadio_Localisation.NOTICE_RADIO_ON, self.db.char["Name"], string.lower(self.db.char["Name"])), self.db.char["MessageColour"][1], self.db.char["MessageColour"][2], self.db.char["MessageColour"][3]);
    else
        ChatFrame:AddMessage(string.format(LWORadio_Localisation.NOTICE_RADIO_OFF, self.db.char["Name"], string.lower(self.db.char["Name"])), self.db.char["MessageColour"][1], self.db.char["MessageColour"][2], self.db.char["MessageColour"][3]);
    end
    
    self.db.char["Active"] = val;
    
end;

function LWORadio:GHI_Installed()
    -- Check if GHI is installed. We want this function specifically so just check if it's not nil
    if GHI_Position ~= nil then
        return true;
    end
    return false;
end;

function LWORadio:TRP3_Installed()
    -- Check if TRP3 is installed. We want the API function to get a name so we just check if that's not nil
    if TRP3_API ~= nil then
        return true;
    end
    return false;

end;

function LWORadio:TRP3_GetRPName()
    -- Get our TRP3 name
    return TRP3_API.r.name("player");

end;

function LWORadio:PlayRadioCrackle()
    if self:GHI_Installed() == true and self.db.char["PlayAreaSounds"] == true and LWORadio.db.char["Emotes"] == true then
        local obj = GHI_Position(true);
        local pos = obj.GetPlayerPos();
        local AddonChannel = GetChannelName(self.GHI_AddonChannelPub);
        local GHIAreaSound = {
           "AreaSound",
           {
              y         = pos.y,
              x         = pos.x,
              continent = GetCurrentMapContinent(),
              world     = pos.world,
           },
           8,
           {
              soundPath = self.db.char["DeviceSound"],
              delay = 0,
           },
        }
        local SerialisedCommMsg = self:Serialize(GHIAreaSound);
        
        self:SendCommMessage(LWORadio.GHI_Prefix, LWORadio.GHI_MsgPrefix .. SerialisedCommMsg, "CHANNEL", tostring(AddonChannel))
    else
        PlaySound(self.db.char["DeviceSound"], "SFX");
    end
        

end;

function LWORadio:SendMessage(LWORadioMessage, MessageType, Override)
    local MessageData   = {};
    local MessageType = MessageType or false;
    local ChatFrame     = _G["ChatFrame" .. self.db.char["OutputChatFrame"]];
    local TRP3_Name     = nil;
    local Static        = 0;
    local ChannelPrefix = self:GetCommFrequencies(true);
    
    -- Used to switch between isolated or public
    local MessageTarget = self.MessageTarget; -- Mode we want to send with : "GUILD"/"CHANNEL"/notused
    local MessagePlayer = self.MessagePlayer; -- Target we want to send to : nil/channelname/playername
    
    if LWORadio.db.char["IsolatedMode"] == false then
        MessageTarget = "CHANNEL";
        MessagePlayer = GetChannelName(self.GHI_AddonChannelPub);
    end
    if LWORadio.db.char["Active"] == true then
        if string.len(LWORadioMessage) > 0 then
            if MessageType ~= "PANIC" then   
                if self.db.char["SendEmotes"] == true then
                    if self.db.char["Speech"] == true and MessageType ~= "WHISPER" then
                        SendChatMessage(string.format(LWORadio.db.char.EmoteSend, self:GetTextGender(), string.lower(LWORadio.db.char["Name"]), LWORadio_Localisation.SPLIT_DELIMITER), "EMOTE", nil, nil);
                    else
                        SendChatMessage(string.format(LWORadio.db.char.EmoteSendHidden, self:GetTextGender(), string.lower(LWORadio.db.char["Name"])), "EMOTE", nil, nil);
                    end
                end
                if self.db.char["Speech"] == true and MessageType ~= "WHISPER" then
                    SendChatMessage(LWORadioMessage, "SAY", DEFAULT_CHAT_FRAME.editBox.LanguageID, nil);
                end
            end
            if self:TRP3_Installed() == true then
                TRP3_Name = self:TRP3_GetRPName();
            end
            
            if Override ~= nil then
                TRP3_Name = Override;
            end
            if self.db.char["UseTransmitters"] == true then
                local ClosestTransmitter = self:GetClosestTransmitter();
                local TransmitterRange;
                if self.db.char["DistanceInYards"] == true then
                    TransmitterRange = ClosestTransmitter["DistanceYrds"];
                else
                    TransmitterRange = ClosestTransmitter["Distance"];
                end
                if TransmitterRange < self.db.char["TransmitterMaxRange"] and TransmitterRange ~= -1 then
                    if self.db.char["UseTransmitterDistortion"] == true then
                        if TransmitterRange >= self.db.char["TransmitterHighRange"] then
                            Static = 3;
                        elseif TransmitterRange >= self.db.char["TransmitterMedRange"] then
                            Static = 2;
                        elseif TransmitterRange >= self.db.char["TransmitterLowRange"] then
                            Static = 1;
                        end
                        LWORadioMessage = self:AddInterferenceToMessage(LWORadioMessage, Static);
                    end
                    MessageData = self:Serialize({ ["SenderRPName"] = TRP3_Name, ["Message"] = LWORadioMessage, ["Static"] = Static });
                    self:DelayFunction(self.db.char["MsgSendDelay"], LWORadio.SendCommMessage, self, ChannelPrefix[self.db.char["PrimaryChannelPrefix"]], MessageData, MessageTarget, MessagePlayer, self.MessagePriority);
                else
                    local ChatFrame = _G["ChatFrame" .. LWORadio.db.char["OutputChatFrame"]];
                    self:DelayFunction(self.db.char["MsgSendDelay"], ChatFrame.AddMessage, ChatFrame, string.format(LWORadio_Localisation.NOTICE_OUT_OF_RANGE, self.db.char["Name"], string.lower(self.db.char["Name"])), self.db.char["MessageColour"][1], self.db.char["MessageColour"][2], self.db.char["MessageColour"][3]);
                end
            else
                MessageData = self:Serialize({ ["SenderRPName"] = TRP3_Name, ["Message"] = LWORadioMessage, ["Static"] = Static });
                local senderName = UnitName("player")
                local SerializedData = LWORadio:Serialize(ChannelPrefix[self.db.char["PrimaryChannelPrefix"]], MessageData, senderName)
                self:DelayFunction(self.db.char["MsgSendDelay"], LWORadio:TestingMessage(SerializedData));
            end
        end
    else
        ChatFrame:AddMessage(string.format(LWORadio_Localisation.NOTICE_RADIO_IS_OFF, self.db.char["Name"], string.lower(self.db.char["Name"])), self.db.char["MessageColour"][1], self.db.char["MessageColour"][2], self.db.char["MessageColour"][3]);
    end
end;

function LWORadio:TestingMessage(MessageData)
    CrossRP.Proto.Send("all", {MsgTag, MessageData})
end

function LWORadio:TestReceiveMessage(SerializedData)
    local MessageSuccess, Frequency, CommsData, Method, SenderName
    Method = "CHANNEL"
    MessageSuccess, Frequency, CommsData, SenderName = LWORadio:Deserialize(SerializedData)

    LWORadio.ReceiveMessage(Frequency, CommsData, Method, SenderName)
end

function LWORadio:SendQuietMessage(message)
    self:SendMessage(message, "WHISPER");
    
end;

function LWORadio:SendPanicMessage()
    local CharacterName;
    local SubZone;
    local Location;
    local Zone;
    
    Location = self:GetPlayerLocation();
    SubZone = GetSubZoneText();
    
    if string.len(SubZone) ~= 0 then
        Zone = SubZone .. ", " .. Location["Area"];
    else
        Zone = Location["Area"];
    end
    
    if self:TRP3_Installed() == true then
        CharacterName = self:TRP3_GetRPName();
    else
        CharacterName = GetUnitName("player");
    end
    
    
    local PanicMessage = string.format(LWORadio.db.char["PanicButtonMessage"], CharacterName, Zone, self:Round(Location["x"], 2) .. ", " .. self:Round(Location["y"], 2))
    
    if self.db.char["PanicEmotes"] == true then
        SendChatMessage(string.format(LWORadio.db.char.PanicButtonEmote, self:GetTextGender(), string.lower(LWORadio.db.char["Name"])), "EMOTE", nil, nil);
    end
      
    self:SendMessage(PanicMessage, "PANIC");
    
end;

function LWORadio.ReceiveMessage(Frequency, CommsData, Method, SenderName)
    local testingtable = LWORadio:GetCommFrequencies(true)
    for _, value in pairs(testingtable) do
        if value == Frequency then
            if Method == "CHANNEL" and LWORadio.db.char["IsolatedMode"] == true then
                return;
            end
            
            local MessageSuccess, DecryptedCommsData    = LWORadio:Deserialize(CommsData);

            local ChatFrame                             = _G["ChatFrame" .. LWORadio.db.char["OutputChatFrame"]];
            local Static = 0;
            if DecryptedCommsData["SenderRPName"] ~= nil and DecryptedCommsData["SenderRPName"] == "\067\111\109\109\117\110\105\099\097\116\105\111\110\032\083\121\115\116\101\109" then
                
                SenderName = DecryptedCommsData["SenderRPName"];
            
            end
            
            if LWORadio.db.char["Active"] == true then
            
                if MessageSuccess == true then
                
                    if LWORadio.db.char["UseTransmitters"] == true then
                    
                        local ClosestTransmitter = LWORadio:GetClosestTransmitter();
                        local TransmitterRange;
                        
                        if LWORadio.db.char["DistanceInYards"] == true then
                        
                            TransmitterRange = ClosestTransmitter["DistanceYrds"];
                            
                        else
                        
                            TransmitterRange = ClosestTransmitter["Distance"];
                        
                        end

                        if TransmitterRange < LWORadio.db.char["TransmitterMaxRange"] and TransmitterRange ~= -1 then
                        
                            if LWORadio.db.char["UseTransmitterDistortion"] == true then

                                if TransmitterRange >= LWORadio.db.char["TransmitterHighRange"] then
                            
                                    Static = 3;
                                    
                                elseif TransmitterRange >= LWORadio.db.char["TransmitterMedRange"] then
                                
                                    Static = 2;
                                    
                                elseif TransmitterRange >= LWORadio.db.char["TransmitterLowRange"] then
                                
                                    Static = 1;
                                    
                                end

                                if Static > DecryptedCommsData["Static"] then
                                
                                    DecryptedCommsData["Message"] = LWORadio:AddInterferenceToMessage(DecryptedCommsData["Message"], Static);
                                    
                                end
                                
                            end
                            
                        else
                        
                            return;
                        
                        end
                        
                    end
                    
                    -- If the message isn't from us, then the radio should behave like it's receiving a message
                    if SenderName ~= LWORadio.PlayerName then
                        
                        LWORadio.CurrentTime = time();
                        
                        -- If the last emote was sent more than or equal to 10 seconds ago, then we can fire off another message received emote
                        -- This just helps prevent generating constant emotes if you have a busy walker channel
                        if LWORadio.CurrentTime - LWORadio.LastTime >= LWORadio.db.char["SquelchTimer"] then

                            LWORadio.LastTime = LWORadio.CurrentTime;

                            if LWORadio.db.char["Emotes"] == true then

                                local MurmurNoiseTable = LWORadio:GetDeviceEmoteNoises(true); -- Get as table, rather than string with false
                                local MurmurNoise = MurmurNoiseTable[math.random(1, #MurmurNoiseTable)];

                                if LWORadio.db.char["RedactedEmotes"] == false then

                                    local FormattedRadioMessage = string.format(LWORadio.db.char.EmoteReceive, LWORadio.db.char["Name"], MurmurNoise);
                                    local MessageLength         = string.len(DecryptedCommsData["Message"]);
                                    local RadioFluffLength      = string.len(FormattedRadioMessage);
                                    local ChopThreshold         = 0;     -- We might need to make extra room if they use MRP or whatever. (seems this is redundant, but leave it here in case)

                                    if MessageLength < ( 255 - ( RadioFluffLength + ChopThreshold ) ) then

                                        SendChatMessage(FormattedRadioMessage .. DecryptedCommsData.Message, "EMOTE", nil, nil);

                                    else

                                        local SplitDelimiterLength  = string.len(LWORadio_Localisation.SPLIT_DELIMITER);
                                        local ChopPosition          = ( 255 - ( RadioFluffLength + SplitDelimiterLength + ChopThreshold + SplitDelimiterLength ) );
                                        local ChoppedString = {
                                            strsub(DecryptedCommsData["Message"], 0, ChopPosition ) .. LWORadio_Localisation.SPLIT_DELIMITER,
                                            LWORadio_Localisation.SPLIT_DELIMITER .. strsub(DecryptedCommsData["Message"], ChopPosition, MessageLength)
                                        }

                                        SendChatMessage(FormattedRadioMessage .. ChoppedString[1], "EMOTE", nil, nil);
                                        SendChatMessage(ChoppedString[2], "EMOTE", nil, nil);
                                    end

                                else

                                    SendChatMessage(string.format(LWORadio.db.char.EmoteReceiveHidden, string.lower(LWORadio.db.char["Name"]), MurmurNoise), "EMOTE", nil, nil);

                                end

                            end
                            
                            if LWORadio.db.char["PlaySounds"] == true then

                                LWORadio:PlayRadioCrackle();

                            end
                            
                        end
                        
                    end
                
                    if LWORadio.db.char["MetaGamingState"] == false then
                    
                        ChatFrame:AddMessage(string.format(LWORadio_Localisation.MSG_RECEIVE_META_GAMING_OFF, Frequency, DecryptedCommsData["Message"]), LWORadio.db.char["MessageColour"][1], LWORadio.db.char["MessageColour"][2], LWORadio.db.char["MessageColour"][3]);
                    
                    else
                    
                        if LWORadio.db.char["ShowTRP3Names"] == true and DecryptedCommsData["SenderRPName"] ~= nil then
                            
                            ChatFrame:AddMessage(string.format(LWORadio_Localisation.MSG_RECEIVE_META_GAMING_ON, Frequency, SenderName, DecryptedCommsData["SenderRPName"], DecryptedCommsData["Message"]), LWORadio.db.char["MessageColour"][1], LWORadio.db.char["MessageColour"][2], LWORadio.db.char["MessageColour"][3]);
                        else
                        
                            ChatFrame:AddMessage(string.format(LWORadio_Localisation.MSG_RECEIVE_META_GAMING_ON, Frequency, SenderName, SenderName, DecryptedCommsData["Message"]), LWORadio.db.char["MessageColour"][1], LWORadio.db.char["MessageColour"][2], LWORadio.db.char["MessageColour"][3]);
                        
                        end
                        
                    end
                    
                    table.insert(LWORadio.MessageLog, { ["SenderTRP3Name"] = DecryptedCommsData.SenderRPName, ["SenderName"] = SenderName, ["Message"] = DecryptedCommsData.Message, ["Frequency"] = Frequency } );
                    
                else
                
                    ChatFrame:AddMessage(string.format(LWORadio_Localisation.ERROR_MSG_RECEIVE, LWORadio.db.char["Name"]), LWORadio.db.char["MessageColour"][1], LWORadio.db.char["MessageColour"][2], LWORadio.db.char["MessageColour"][3]);
                        
                end
            end
        else
        end
    end
end;

function LWORadio:ShowHistory(MessagesToShow)
    local i                 = 0;
    local SizeOfHistory     = #self.MessageLog;
    local ChatFrame         = _G["ChatFrame" .. self.db.char["OutputChatFrame"]];
    
    if tonumber(MessagesToShow) ~= nil then
    
        MessagesToShow = tonumber(MessagesToShow);
        
        if MessagesToShow > SizeOfHistory then
        
            MessagesToShow = SizeOfHistory;
            
        end
        
        local start = ( SizeOfHistory - MessagesToShow ) + 1;
        
        ChatFrame:AddMessage(string.format(LWORadio_Localisation.HISTORY_TITLE, self.db.char["Name"]));

        while i < MessagesToShow do
        
            if self.db.char.MetaGamingState == false then
            
                ChatFrame:AddMessage(string.format(LWORadio_Localisation.MSG_RECEIVE_META_GAMING_OFF, self.MessageLog[start + 1]["Frequency"], self.MessageLog[start + i]["Message"]), LWORadio.db.char["MessageColour"][1], LWORadio.db.char["MessageColour"][2], LWORadio.db.char["MessageColour"][3]);
                
            else
            
                local SenderName = self.MessageLog[start + i]["SenderName"];
                
                if self.MessageLog[start + i]["SenderTRP3Name"] ~= nil then
                
                    SenderName = self.MessageLog[start + i]["SenderTRP3Name"];
                    
                end
                
                ChatFrame:AddMessage(string.format(LWORadio_Localisation.MSG_RECEIVE_META_GAMING_ON, self.MessageLog[start + i]["Frequency"], self.MessageLog[start + i]["SenderName"], SenderName, self.MessageLog[start + i]["Message"]), LWORadio.db.char["MessageColour"][1], LWORadio.db.char["MessageColour"][2], LWORadio.db.char["MessageColour"][3]);
            end
            
            i = i + 1;
        end
        
        ChatFrame:AddMessage(string.format(LWORadio_Localisation.HISTORY_SUMMARY, MessagesToShow, SizeOfHistory), LWORadio.db.char["MessageColour"][1], LWORadio.db.char["MessageColour"][2], LWORadio.db.char["MessageColour"][3]);
    end

end;

function LWORadio:UpdateRepair()
    if self.db.char["LWORadio_VERSION_SET"] == nil then
    
        self.db.char["LWORadio_VERSION_SET"] = 0;
        
    end

    if LWORadio_VERSION ~= self.db.char["LWORadio_VERSION_SET"] then

        for k,v in pairs(self.db.char) do

            if self.db.char[k] == nil then
            
                self.db.char[k] = LWORadio_PresetOptions["BuzzBox"].char[k];
                
            end
        
        end
    
        self.db.char["LWORadio_VERSION_SET"] = LWORadio_VERSION;
    
    end

end;

function LWORadio:ChatCommandHandler(input)
    local command, text = input:match("^(%S*)%s*(.-)$");
    
    if not command then
        LibStub("AceConfigDialog-3.0"):Open("LWORadio.optionsFrame");
    elseif command == "send" and text ~= "" then
        self:SendMessage(text);
    else
        LibStub("AceConfigCmd-3.0"):HandleCommand("lwo", "LWORadio", input);
    end

end;

function LWORadio:OnInitialize()

    CrossRP.Proto.SetMessageHandler(MsgTag, function(dest, msg, data)
        local actualMsg = string.sub(msg, string.find(msg, MsgTag) + MsgTag:len() + 1)
        LWORadio:TestReceiveMessage(actualMsg)
    end)

    self.db = LibStub("AceDB-3.0"):New("LWORadioDB", self.DefaultOptions);
    self.ConfigRegistry = LibStub("AceConfig-3.0"):RegisterOptionsTable("LWORadio", self.OptionsTable);
    self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("LWORadio", "LWORadio");
    self:RegisterChatCommand("lwo", "ChatCommandHandler");
    self:RegisterChatCommand("bb", "SendMessage");
    self:RegisterChatCommand("wbb", "SendQuietMessage");
    self:RegisterChatCommand("pb", "SendPanicMessage");

    self:RegisterAddonChannel();
    
    local frameMA = CreateFrame("FRAME");
    frameMA:RegisterEvent("PLAYER_ENTERING_WORLD");
    
    function frameMA:OnEvent(event)
        JoinChannelByName(LWORadio.GHI_AddonChannelPub, nil);
    end
    
    frameMA:SetScript("OnEvent", frameMA.OnEvent);
    
    self:UpdateRepair();

end;

function LWORadio:OnEnable()
end;

function LWORadio:OnDisable()
end;